#ifndef AudioDeviceInterface_p_h
#define AudioDeviceInterface_p_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CddbQuery.h"
#include "Counter.h"
#include "NonCopyable.h"
#include "ObjectDeleter.h"
#include "TrackInformation.h"

#include <QObject>
#include <QTcpServer>
#include <QUrl>

#if WITH_CDIO_PARANOIA
#include <cdio/cdda.h>
#include <cdio/paranoia.h>
#endif

#if WITH_CDDB
#include <cddb/cddb.h>
#endif

#include <unistd.h>
#include <memory>


//* cddb disc info
class CddbDiscInfo
{
    public:

    //* clear
    void clear()
    { *this = CddbDiscInfo(); }

    //* first track
    int firstTrack_ = 0;
    int lastTrack_ = 0;

    // contains lastTrack - firstTrack + 2 entries
    // last entry corresponds to disc length
    QList<int> frameOffsets_;
};

//* audio device private interface
class AudioDeviceInterface::Private: private Base::Counter<AudioDeviceInterface::Private>
{

    public:

    //* constructor
    Private():
        Counter( "AudioDeviceInterface::Private" )
    {}

    //* setup server
    void setupServer();

    //*@name accessors
    //@{

    //* create QUrl from track information
    QUrl url( const TrackInformation& ) const;

    //@}

    //*@name modifiers
    //@{

    //* create track list from CddbDisc info
    void createTrackList();

    //* create CddbDisc from  CddbDisc info
    void createCddbDisc();

    //@}

    //* Server
    std::unique_ptr<QTcpServer, Base::ObjectDeleter> server_;

    //* server port
    int port_ = 0;

    //* device
    QString device_;

    //* tracks
    TrackInformation::List trackList_;

    //* cddb disc info
    CddbDiscInfo cddbDiscInfo_;

    //* elapsed time
    quint32 elapsedTime_ = 0;

    #if WITH_CDIO_PARANOIA
    class DriveDeleter
    {
        public:
        void operator()( cdrom_drive_t* drive ) const
        { cdio_cddap_close(drive); }

    };

    //* cdio drive
    std::unique_ptr<cdrom_drive_t, DriveDeleter> drive_;

    class ParanoiaDeleter
    {
        public:
        void operator()( cdrom_paranoia_t* paranoia ) const
        { cdio_paranoia_free( paranoia ); }
    };

    //* cdio paranoia drive
    std::unique_ptr<cdrom_paranoia_t, ParanoiaDeleter> paranoia_;
    #endif

    #if WITH_CDDB
    //* cddb disc
    std::shared_ptr<cddb_disc_t> cddbDisc_;
    #endif

};

//* file descriptor with auto-close
class FileDescriptor: private Base::NonCopyable<FileDescriptor>
{
    public:

    //* constructor
    explicit FileDescriptor( int fd ):
        fd_(fd)
        {}

    //* destructor
    ~FileDescriptor()
    { if( fd_ >= 0 ) close( fd_ ); }

    //* cast
    operator const int& () const
    { return fd_; }

    private:

    //* actual file descriptor
    int fd_ = 0;

};

#endif
