#ifndef CollectionWidget_h
#define CollectionWidget_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BusyWidget.h"
#include "Counter.h"
#include "CollectionItem.h"
#include "CollectionModel.h"

#include <QAction>
#include <QWidget>

class ScrollBarMonitor;
class CollectionView;
class CollectionFindWidget;
class ToolTipWidget;

class CollectionWidget: public QWidget, private Base::Counter<CollectionWidget>
{

    Q_OBJECT

    public:

    //* constructor
    explicit CollectionWidget( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* collection find widget
    CollectionFindWidget& findWidget() const
    { return *findWidget_; }

    //* list
    CollectionView& list() const
    { return *list_; }

    //* model
    const CollectionModel& model() const
    { return model_; }

    //* list of expanded items
    CollectionItem::List expandedItems() const;

    //* tooltip widget
    ToolTipWidget& toolTipWidget() const
    { return *toolTipWidget_; }

    //@}

    //*@name actions
    //@{

    QAction& updateCollectionAction() const
    { return *updateCollectionAction_; }

    //@}

    //*@name modifiers
    //@{

    //* model
    CollectionModel& model()
    { return model_; }

    //* assign albums to artists
    /** this requires that items are sorted */
    void updateArtists( const CollectionItem::List& );

    //* assign tracks to albums
    /** this requires that items are sorted */
    void updateAlbums( const CollectionItem::List& );

    //* save expanded items
    void saveExpandedItems();

    //* restore
    void restoreExpandedItems();

    //* lock
    void lock();

    //* unlock
    void unlock();

    //@}

    Q_SIGNALS:

    //* request collection update
    void requestCollectionUpdate();

    //* request albums matching artist
    void requestAlbums( const CollectionItem& );

    //* request tracks matching artist and album
    void requestTracks( const CollectionItem& );

    //* add items to playlist
    void requestAddToPlaylist( const CollectionItem::List& );

    //* replace playlist
    void requestReplacePlaylist( const CollectionItem::List& );

    //* delete tracks
    void requestDeleteTracks( const CollectionItem::List& );

    #if WITH_TAGLIB
    //* request collection update
    void requestEditTags( const CollectionItem::List& );
    #endif

    public Q_SLOTS:

    //* busy
    void busy()
    {
        isBusy_ = true;
        busyWidget_->start();
    }

    //* busy
    void idle()
    {
        isBusy_ = false;
        busyWidget_->stop();
    }

    //* toggle busy
    void toggleBusy()
    {
        if( isBusy_ )
        {
            isBusy_ = false;
            busyWidget_->stop();
        } else {
            isBusy_ = true;
            busyWidget_->start();
        }
    }

    private Q_SLOTS:

    //* index activated
    void _itemActivated( const QModelIndex& );

    //* index exmapnded
    void _itemExpanded( const QModelIndex& );

    //* update actions
    void _updateActions();

    //* add to playlist
    void _addSelectionToPlaylist();

    //* replace playlist
    void _replacePlaylist();

    //* delete tracks
    void _deleteTracks();

    //* show tooltip
    void _showToolTip( const QModelIndex& );

    #if WITH_TAGLIB
    //* edit tags
    void _editTags();
    #endif

    private:

    //* actions
    void _installActions();

    //* locked
    /** prevents roundtrip to mpd */
    bool locked_ = false;

    //* collection find widget
    CollectionFindWidget* findWidget_ = nullptr;

    //* treeview
    CollectionView* list_ = nullptr;

    //* scrollbar monitor
    ScrollBarMonitor* scrollBarMonitor_ = nullptr;

    //* model
    CollectionModel model_;

    //* expanded items
    CollectionItem::List expandedItems_;

    //* tooltip widget
    ToolTipWidget* toolTipWidget_ = nullptr;

    //* busy widget
    BusyWidget* busyWidget_ = nullptr;

    //* true if busy
    bool isBusy_ = false;

    //*@name actions
    //@{

    //* update collection
    QAction* updateCollectionAction_ = nullptr;

    //* add to playlist
    QAction* addToPlaylistAction_ = nullptr;

    //* replace playlist
    QAction* replacePlaylistAction_ = nullptr;

    //* delete tracks
    QAction* deleteTracksAction_ = nullptr;

    #if WITH_TAGLIB
    //* edit tags
    QAction* editTagsAction_ = nullptr;
    #endif

    //@}
};

#endif
