/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConfigurationDialog.h"

#include "AudioDeviceSpeed.h"
#include "CacheInformationWidget.h"
#include "CppUtil.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "LyricsProviderWidget.h"
#include "OptionBrowsedLineEditor.h"
#include "OptionCheckBox.h"
#include "OptionComboBox.h"
#include "OptionIconSizeComboBox.h"
#include "OptionLineEditor.h"
#include "OptionSpinBox.h"

#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QPushButton>
#include <QSystemTrayIcon>

//_________________________________________________________
ConfigurationDialog::ConfigurationDialog( QWidget* parent ):
    BaseConfigurationDialog( parent )
{}

//_________________________________________________________
void ConfigurationDialog::realizeWidget()
{

    // base configuration
    baseConfiguration( nullptr, Flag::Base|Flag::List);

    // player configuration
    auto page = &addPage( IconEngine::get( IconNames::PreferencesGeneral ), tr( "Playback" ), tr( "MPD player settings" ) );
    {
        QWidget* box = new QWidget( page );
        page->layout()->addWidget( box );

        auto vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        auto hLayout = new QHBoxLayout;
        hLayout->setSpacing(0);
        vLayout->addLayout( hLayout );

        QLabel* label;
        OptionBrowsedLineEditor* editor;
        hLayout->addWidget( label = new QLabel( tr( "Mpd command: " ), box ) );
        hLayout->addWidget( editor = new OptionBrowsedLineEditor( box, "MPD_COMMAND" ) );
        addOptionWidget( editor );
        label->setBuddy( editor );
        editor->setTargetApplication( File("mpd") );

        box = new QGroupBox( tr( "Options" ), page );
        page->layout()->addWidget( box );
        vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        OptionCheckBox* checkbox;

        #if !defined(Q_OS_WIN)
        // no pulse audio under windows
        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use pulse audio sound server" ), box, "MPD_USE_PULSE_AUDIO" ) );
        addOptionWidget( checkbox );

        // we force using tcp sockets on windows. Local sockets do not work
        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use tcp sockets to communicate with mpd" ), box, "MPD_USE_TCP_SOCKET" ) );
        addOptionWidget( checkbox );
        #endif

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Resume playback on start" ), box, "RESUME_PLAYBACK_ON_START" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Close mpd daemon on exit" ), box, "MPD_CLOSE_DEAMON_ON_EXIT" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Enable log" ), box, "LOG_ENABLED" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Fadeout on stop" ), box, "USE_FADE_ANIMATION" ) );
        addOptionWidget( checkbox );

        hLayout = new QHBoxLayout;
        hLayout->setMargin(0);
        box->layout()->addItem( hLayout );
        hLayout->addWidget( label = new QLabel( tr( "Fadeout duration: " ), box ) );
        label->setEnabled(false);

        OptionSpinBox* spinbox;
        hLayout->addWidget( spinbox = new OptionSpinBox( box, "FADE_ANIMATION_DURATION" ) );
        spinbox->setMinimum( 500 );
        spinbox->setMaximum( 10000 );
        spinbox->setSingleStep( 500 );
        spinbox->setSuffix( tr( " ms" ) );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );
        spinbox->setEnabled(false);

        connect( checkbox, SIGNAL(toggled(bool)), label, SLOT(setEnabled(bool)) );
        connect( checkbox, SIGNAL(toggled(bool)), spinbox, SLOT(setEnabled(bool)) );

    }

    // appearance
    page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Appearance" ), tr( "Application's appearance" ) );
    {

        auto box = new QWidget( page );
        page->layout()->addWidget( box );

        auto vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        OptionCheckBox* checkbox = nullptr;
        vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Show stop button" ), box, "SHOW_STOP_BUTTON" ) );
        addOptionWidget( checkbox );

        vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Show artist icon in collection view" ), box, "SHOW_ARTIST_ICON" ) );
        addOptionWidget( checkbox );

        vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Show tooltips" ), box, "SHOW_TOOLTIPS" ) );
        addOptionWidget( checkbox );

        // icon size in lists
        auto gridLayout = new GridLayout;
        gridLayout->setMargin(0);
        gridLayout->setMaxCount(3);
        gridLayout->setColumnStretch( 2, 1 );
        vLayout->addLayout( gridLayout );
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );

        QLabel* label;
        {
            // icon size in playlist
            auto comboBox = new OptionIconSizeComboBox( box, "PLAYLIST_ICON_SIZE" );
            gridLayout->addWidget( label = new QLabel( tr( "Playlist icon size:" ), box ) );
            gridLayout->addWidget( comboBox );
            addOptionWidget( comboBox );
            label->setBuddy( comboBox );
            gridLayout->increment();
        }

        {
            // icon size in tooltip
            auto comboBox = new OptionIconSizeComboBox( box, "TOOLTIPS_PIXMAP_SIZE" );
            gridLayout->addWidget( label = new QLabel( tr( "Tooltips icon size:" ), box ) );
            gridLayout->addWidget( comboBox );
            addOptionWidget( comboBox );
            label->setBuddy( comboBox );
            gridLayout->increment();
        }

    }

    // behavior
    page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Behavior" ), tr( "Application's behavior" ) );
    {
        QWidget* box = new QGroupBox( tr( "System Tray" ), page );
        page->layout()->addWidget( box );

        auto vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        bool hasSystemTray( QSystemTrayIcon::isSystemTrayAvailable() );

        OptionCheckBox* sysTrayCheckBox = nullptr;
        vLayout->addWidget( sysTrayCheckBox = new OptionCheckBox( tr( "Use system tray" ), box, "USE_SYSTEMTRAY" ) );
        addOptionWidget( sysTrayCheckBox );

        OptionCheckBox* checkbox = nullptr;
        vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Use monochrome system tray icon" ), box, "USE_MONOCHROME_SYSTEMTRAY_ICON" ) );
        addOptionWidget( checkbox );
        connect( sysTrayCheckBox, SIGNAL(toggled(bool)), checkbox, SLOT(setEnabled(bool)) );

        auto hLayout = new QHBoxLayout();
        hLayout->setMargin(0);
        vLayout->addLayout( hLayout );

        QLabel* label;
        hLayout->addWidget( label = new QLabel( tr( "Window visibility on startup: " ), box ) );

        OptionComboBox* comboBox;
        hLayout->addWidget( comboBox = new OptionComboBox( box, "MAIN_WINDOW_VISIBILITY" ) );
        comboBox->setUseValue( false );
        comboBox->addItems( Base::makeT<QStringList>( { tr( "Show window on startup" ), tr( "Hide window on startup" ), tr( "Restore previous state" ) } ) );
        label->setBuddy( comboBox );
        addOptionWidget( comboBox );

        connect( sysTrayCheckBox, SIGNAL(toggled(bool)), label, SLOT(setEnabled(bool)) );
        connect( sysTrayCheckBox, SIGNAL(toggled(bool)), comboBox, SLOT(setEnabled(bool)) );
        connect( sysTrayCheckBox, SIGNAL(toggled(bool)), this, SLOT(_toggleTextNotifications(bool)) );

        box->setEnabled( hasSystemTray );

        box = new QWidget( page );
        page->layout()->addWidget( box );

        vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        // collapse/expand on double click
        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Collapse/expand main window on double-click" ), box, "COLLAPSE_ON_DOUBLE_CLICK" ) );
        addOptionWidget( checkbox );

        // notifications
        box->layout()->addWidget( textNotificationsCheckBox_ = new OptionCheckBox( tr( "Use text notifications" ), box, "USE_NOTIFICATIONS" ) );
        addOptionWidget( textNotificationsCheckBox_ );
        textNotificationsCheckBox_->setEnabled( textNotificationsSupported_ );

        #if WITH_DBUS
        // dbus interface
        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Enable MPRIS D-Bus interface" ), box, "USE_DBUS_INTERFACE" ) );
        addOptionWidget( checkbox );
        #endif

    }

    // collection
    page = &addPage( IconEngine::get( IconNames::PreferencesCollection ), tr( "Collection" ), tr( "Music collection settings" ) );
    {
        QWidget* box = new QWidget( page );
        page->layout()->addWidget( box );

        auto vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        auto gridLayout = new GridLayout;
        gridLayout->setMargin(0);
        gridLayout->setMaxCount(3);
        vLayout->addLayout( gridLayout );

        QLabel* label;
        gridLayout->addWidget( label = new QLabel( tr( "Music folder:" ), box ) );
        label->setAlignment( Qt::AlignVCenter|Qt::AlignRight );

        OptionBrowsedLineEditor* editor;
        gridLayout->addWidget( editor = new OptionBrowsedLineEditor( box, "MUSIC_DIR" ), 0, 1, 1, 2 );
        editor->setFileMode( QFileDialog::Directory );
        label->setBuddy( editor );
        addOptionWidget( editor );

        QPushButton* button;
        gridLayout->increment();
        gridLayout->addWidget( button = new QPushButton( IconEngine::get( IconNames::Reload ), tr( "Full Rescan" ), box ) );
        button->setToolTip( tr( "Rescan the entire collection." ) );
        connect( button, SIGNAL(clicked()), this, SIGNAL(requestCollectionUpdate()) );

        gridLayout->setColumnStretch( 2, 1 );

        box = new QGroupBox( tr( "Options" ), page );
        page->layout()->addWidget( box );
        vLayout = new QVBoxLayout;
        box->setLayout( vLayout );

        OptionCheckBox* checkBox;
        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Update collection at startup" ), box, "UPDATE_COLLECTION_AT_STARTUP" ) );
        addOptionWidget( checkBox );

        vLayout->addWidget( checkBox = new OptionCheckBox( tr( "Watch folder for changes" ), box, "WATCH_COLLECTION_CHANGES" ) );
        addOptionWidget( checkBox );

    }

    // audio cd
    page = &addPage( IconEngine::get( IconNames::PreferencesAudioDevice ), tr( "Audio CD" ), tr( "Audio CD settings" ) );
    {
        QWidget* box = new QGroupBox( tr( "Audio CD device" ), page );
        page->layout()->addWidget( box );

        auto gridLayout = new GridLayout;
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter);
        box->setLayout( gridLayout );

        QLabel* label;
        OptionBrowsedLineEditor* browseEditor;
        gridLayout->addWidget( label = new QLabel( tr( "Audio device:" ), box ) );
        gridLayout->addWidget( browseEditor = new OptionBrowsedLineEditor( box, "AUDIO_DEVICE" ) );
        browseEditor->setFileMode( QFileDialog::ExistingFile );
        browseEditor->setAcceptMode( QFileDialog::AcceptOpen );
        label->setBuddy( browseEditor );
        addOptionWidget( browseEditor );

        OptionComboBox* comboBox;
        gridLayout->addWidget( label = new QLabel( tr( "Audio device speed:" ), box ) );
        gridLayout->addWidget( comboBox = new OptionComboBox( box, "AUDIO_DEVICE_SPEED" ) );
        comboBox->addItems( AudioDeviceSpeed::speedStringList() );
        comboBox->setUseValue( false );
        label->setBuddy( comboBox );
        addOptionWidget( comboBox );

        box = new QGroupBox( tr( "Album and Track information retrieval" ), page );
        page->layout()->addWidget( box );

        gridLayout = new GridLayout;
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter);
        box->setLayout( gridLayout );

        OptionLineEditor* editor;
        gridLayout->addWidget( label = new QLabel( tr( "CDDB host:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "CDDB_HOST" ) );
        label->setBuddy( editor );
        addOptionWidget( editor );

        OptionSpinBox* spinbox;
        gridLayout->addWidget( label = new QLabel( tr( "CDDB port:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "CDDB_PORT" ) );
        spinbox->setMinimum( 1 );
        spinbox->setMaximum( 10000 );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

    }

    // lyrics
    page = &addPage( IconEngine::get( IconNames::Lyrics ), tr( "Lyrics" ), tr( "Lyrics providers" ) );
    {
        page->layout()->addWidget( lyricsProviderWidget_ = new LyricsProviderWidget( page ) );
    }

    // cache
    page = &addPage( IconEngine::get( IconNames::PreferencesLocalCache ), tr( "Cache" ), tr( "Cached items" ) );
    {
        page->layout()->addWidget( new CacheInformationWidget( page ) );
    }

    // load initial configuration
    read();

}

//________________________________________________________________________
void ConfigurationDialog::_toggleTextNotifications( bool value )
{ textNotificationsCheckBox_->setEnabled( textNotificationsSupported_ || ( value &&  QSystemTrayIcon::supportsMessages() && QSystemTrayIcon::isSystemTrayAvailable() ) ); }
