/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CoverManager.h"
#include "CoverManager_p.h"
#include "CppUtil.h"
#include "Debug.h"
#include "QUrlUtil.h"
#include "XmlError.h"
#include "XmlOptions.h"

#include <QDir>
#include <QImage>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QUrl>

#include <QDomDocument>

namespace Private
{
    //_______________________________________________________________________________
    CoverJob::CoverJob( QNetworkReply* networkReply ):
        Counter( "Private::CoverJob" ),
        networkReply_( networkReply )
    {
        Debug::Throw( "Private::CoverJob::CoverJob.\n" );

        // connect
        connect( networkReply_.get(), SIGNAL(finished()), this, SLOT(_finished()) );
        connect( networkReply_.get(), SIGNAL(error(QNetworkReply::NetworkError)), this, SLOT(_abort()) );
    }

    //_______________________________________________________________________________
    CoverJob::~CoverJob() = default;

    //_______________________________________________________________________________
    void CoverJob::_abort()
    {
        Debug::Throw( "Private::CoverJob::_abort.\n" );
        networkReply_.reset();
        emit finished();
    }

    //_______________________________________________________________________________
    CoverLocationJob::CoverLocationJob( const CollectionItem& collectionItem, QNetworkReply* networkReply ):
        CoverJob( networkReply ),
        collectionItem_( collectionItem )
    { Debug::Throw( "Private::CoverLocationJob::CoverLocationJob.\n" ); }

    //_______________________________________________________________________________
    void CoverLocationJob::_finished()
    {
        Debug::Throw()
            << "Private::CoverLocationJob::_finished -"
            << " artist: " << collectionItem_.artist()
            << " album: " << collectionItem_.album()
            << endl;

        // make sure finished signal is sent when exiting this method
        class SignalEmiter
        {
            public:

            //*constructor
            SignalEmiter( CoverLocationJob* job ):
                job_( job )
                {}

            //* destructor
            ~SignalEmiter()
            { emit job_->finished(); }

            private:

            //* job
            CoverLocationJob* job_;

        };

        SignalEmiter( this );

        // check reply and status
        if( !( _hasNetworkReply() && _networkReply().error() == QNetworkReply::NoError ) )
        { return; }

        const bool isArtist( collectionItem_.type() == CollectionItem::Type::Artist );

        QString extraLargeUrl;
        QString largeUrl;

        // read data
        const auto data( _networkReply().readAll() );

        // create dom document
        QDomDocument document;
        XmlError error;
        if( !document.setContent( data, error ) ) return;

        // parse document
        const auto nodeList( document.elementsByTagName( "image" ) );
        for( int i = 0; i < nodeList.count(); ++i )
        {

            // get element
            const auto element( nodeList.at( i ).toElement() );

            // check parent
            if( element.parentNode().nodeName() != ( isArtist ? QLatin1String( "artist"):QLatin1String("album" ) ) ) continue;

            // get size key, url, and assign
            auto size( element.attribute("size") );
            auto url( element.text() );
            if( size == QLatin1String("extralarge") ) extraLargeUrl = url;
            else if( size == QLatin1String("large") ) largeUrl = url;

            // break here if both urls are found
            if( !( extraLargeUrl.isEmpty() || largeUrl.isEmpty() ) ) break;
        }

        if( !extraLargeUrl.isEmpty() ) emit coverFound( collectionItem_, QUrl( extraLargeUrl ) );
        else if( !largeUrl.isEmpty() ) emit coverFound( collectionItem_, QUrl( largeUrl ) );

    }

    //_______________________________________________________________________________
    CoverDownloadJob::CoverDownloadJob( const CollectionItem& collectionItem, QNetworkReply* networkReply ):
        CoverJob( networkReply ),
        collectionItem_( collectionItem )
    { Debug::Throw( "Private::CoverDownloadJob::CoverDownloadJob.\n" ); }

    //_______________________________________________________________________________
    void CoverDownloadJob::_finished()
    {
        Debug::Throw( "Private::CoverDownloadJob::_finished.\n" );

        // check reply and status
        if( _hasNetworkReply() && _networkReply().error() == QNetworkReply::NoError )
        { emit coverFound( collectionItem_, QImage::fromData( _networkReply().readAll() ) ); }

        emit finished();

    }

}

//_______________________________________________________________________________
CoverManager::CoverManager( QObject* parent ):
    QObject( parent ),
    Counter( "CoverManager" ),
    networkAccessManager_( new QNetworkAccessManager( this ) )
{

    Debug::Throw( "CoverManager::CoverManager.\n" );

    // make sure cover path exists
    QDir().mkpath( XmlOptions::get().raw( "ALBUM_ART_CACHE_DIR" ) );

}

//_______________________________________________________________________________
File CoverManager::fileName( const CollectionItem& collectionItem )
{ return File( "AlbumArt_{" + QString::number( collectionItem.hashKey() ) + "}.png" ).addPath( File( XmlOptions::get().raw( "ALBUM_ART_CACHE_DIR" ) ) ); }

//_______________________________________________________________________________
void CoverManager::get( const CollectionItem& collectionItem )
{

    Debug::Throw( "CoverManager::get.\n" );

    // check collection item
    if( collectionItem.artist().isEmpty() ) return;
    if( collectionItem.type() == CollectionItem::Type::Album && collectionItem.album().isEmpty() ) return;

    // try load from disc
    if( _fromDisk( collectionItem ) ) return;
    else {

        // try load from last.fm
        const bool isArtist( collectionItem.type() == CollectionItem::Type::Artist );

        // setup query items
        auto queryItems = Base::makeT<Base::QueryItemList>(
        {
            { QLatin1String( "method" ), isArtist ? QLatin1String( "artist.getInfo" ) : QLatin1String( "album.getInfo" ) },
            { QLatin1String( "api_key" ), QLatin1String( "5a854b839b10f8d46e630e8287c2299b" ) },
            { QLatin1String( "autocorrect" ), QLatin1String( "1" ) },
            { QLatin1String( "artist" ), collectionItem.artist() }
        });

        if( !isArtist ) queryItems.append( { QLatin1String( "album" ), collectionItem.album() } );

        // create url
        QUrl url("https://ws.audioscrobbler.com/2.0/");
        QUrlUtil::addQueryItems( url, queryItems );

        // network reply
        auto coverLocationJob = new Private::CoverLocationJob( collectionItem, networkAccessManager_->get( QNetworkRequest(url) ) );
        connect( coverLocationJob, SIGNAL(coverFound(CollectionItem,QUrl)), SLOT(_download(CollectionItem,QUrl)) );
        connect( coverLocationJob, SIGNAL(finished()), coverLocationJob, SLOT(deleteLater()) );
    }

}


//_______________________________________________________________________________
void CoverManager::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();
        if( !pendingCovers_.empty() )
        {
            emit coversFound( pendingCovers_ );
            pendingCovers_.clear();
        }

    } else return QObject::timerEvent( event );

}

//_______________________________________________________________________________
void CoverManager::_download( const CollectionItem& collectionItem, const QUrl& source )
{
    Debug::Throw( "CoverManager::_download.\n" );
    auto coverDownloadJob = new Private::CoverDownloadJob( collectionItem, networkAccessManager_->get( QNetworkRequest( source ) ) );
    connect( coverDownloadJob, SIGNAL(coverFound(CollectionItem,QImage)), SLOT(_save(CollectionItem,QImage)) );
    connect( coverDownloadJob, SIGNAL(finished()), coverDownloadJob, SLOT(deleteLater()) );
}

//_______________________________________________________________________________
void CoverManager::_save( const CollectionItem& collectionItem, const QImage& image )
{
    Debug::Throw( "CoverManager::_save.\n" );

    if( !image.isNull() )
    {

        // save to disk
        image.save( fileName( collectionItem ) );

        // emit signal
        _delayedCoverFound( collectionItem, image );
    }

}

//_______________________________________________________________________________
bool CoverManager::_fromDisk( const CollectionItem& collectionItem )
{

    Debug::Throw( "CoverManager::_fromDisk.\n" );
    QImage image( fileName( collectionItem ) );
    if( !image.isNull() )
    {

        _delayedCoverFound( collectionItem, image );
        return true;

    } else return false;

}

//_______________________________________________________________________________
void CoverManager::_delayedCoverFound( const CollectionItem& collectionItem, const QImage& image )
{
    pendingCovers_.append( { collectionItem, image } );
    if( !timer_.isActive() ) timer_.start( 100, this );
}
