/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "LyricsProviderWikia.h"
#include "BaseLyricsProvider_p.h"

#include "CppUtil.h"
#include "Debug.h"
#include "QUrlUtil.h"
#include "XmlError.h"

#include <QDomDocument>
#include <QNetworkRequest>
#include <QUrl>

//_____________________________________________________
LyricsProviderWikia::LyricsProviderWikia( QObject* parent ):
    BaseLyricsProvider( parent )
{
    Debug::Throw( "LyricsProviderWikia::LyricsProviderWikia.\n" );
    setName( QLatin1String( "lyrics.wikia.com" ) );
}

//_____________________________________________________
void LyricsProviderWikia::get( const TrackInformation& trackInformation )
{

    Debug::Throw( "LyricsProviderWikia::get.\n" );

    // check network access manager
    if( !_hasNetworkAccessManager() )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // construct url
    QUrl url("http://lyrics.wikia.com/api.php");

    const auto queryItems = Base::makeT<Base::QueryItemList>(
    {
        { QLatin1String("artist"), trackInformation.artist() },
        { QLatin1String("song"), trackInformation.title() },
        { QLatin1String("func"), QLatin1String("getSong") },
        { QLatin1String("fmt"), QLatin1String("xml") }
    } );

    QUrlUtil::addQueryItems( url, queryItems );

    // send request
    auto lyricsRequest = new Private::LyricsRequest( trackInformation, _networkAccessManager().get( QNetworkRequest(url) ) );
    connect( lyricsRequest, SIGNAL(lyricsFound(TrackInformation,QByteArray)), SLOT(_processLyricsLocation(TrackInformation,QByteArray)));
    connect( lyricsRequest, SIGNAL(lyricsNotFound(TrackInformation)), SIGNAL(lyricsNotFound(TrackInformation)));
    connect( lyricsRequest, SIGNAL(finished()), lyricsRequest, SLOT(deleteLater()) );

}

//_____________________________________________________
void LyricsProviderWikia::_processLyricsLocation( const TrackInformation& trackInformation, const QByteArray& data )
{
    Debug::Throw() << "LyricsProviderWikia::_processLyricsLocation -"
        << " track; " << trackInformation << endl
        << " reply: " << data << endl;

    // create dom document
    QDomDocument document;
    XmlError error;
    if( !document.setContent( data, error ) )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // parse document
    const auto nodeList( document.elementsByTagName( "url" ) );
    QString urlString;
    for( int i = 0; i < nodeList.count(); ++i )
    {
        // get element
        const auto element( nodeList.at( i ).toElement() );

        // check parent
        if(
            element.parentNode().nodeName() == QLatin1String( "LyricsResult" ) &&
            !(urlString = element.text() ).isEmpty() &&
            !urlString.contains( "action==edit" ) )
        { break; }

    }

    // find path from url
    auto path( QUrl::fromEncoded(urlString.toUtf8()).path().replace( '+', "%2b" ) );
    if( path.startsWith( '/' ) ) path = path.mid(1);
    path = QUrl::toPercentEncoding( path );

    // construct query url
    QUrl url("http://lyrics.wikia.com/api.php");
    const auto queryItems = Base::makeT<Base::QueryItemList>(
    {
        { QLatin1String("action"), QLatin1String("query") },
        { QLatin1String("prop"), QLatin1String("revisions") },
        { QLatin1String("rvprop"), QLatin1String("content") },
        { QLatin1String("format"), QLatin1String("xml") },
        { QLatin1String("titles"), path }
    });

    QUrlUtil::addQueryItems( url, queryItems );

    // send request
    auto lyricsRequest = new Private::LyricsRequest( trackInformation, _networkAccessManager().get( QNetworkRequest(url) ) );
    connect( lyricsRequest, SIGNAL(lyricsFound(TrackInformation,QByteArray)), SLOT(_processLyrics(TrackInformation,QByteArray)));
    connect( lyricsRequest, SIGNAL(lyricsNotFound(TrackInformation)), SIGNAL(lyricsNotFound(TrackInformation)));
    connect( lyricsRequest, SIGNAL(finished()), lyricsRequest, SLOT(deleteLater()) );

}

//_____________________________________________________
void LyricsProviderWikia::_processLyrics( const TrackInformation& trackInformation, const QByteArray& data )
{
    Debug::Throw() << "LyricsProviderWikia::_processLyrics -"
        << " track; " << trackInformation << endl
        << " reply: " << data << endl;

    QDomDocument document;
    XmlError error;
    if( !document.setContent( data, error ) )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // parse document
    const auto nodeList( document.elementsByTagName( "rev" ) );
    QString lyrics;
    for( int i = 0; i < nodeList.count(); ++i )
    {

        // get element
        const auto element( nodeList.at( i ).toElement() );

        // check parent
        if( element.parentNode().nodeName() == QLatin1String( "revisions" ) )
        { lyrics = element.text(); }

    }

    lyrics = _extract( lyrics, "<lyrics>", "</lyrics>" );

    if( lyrics.isEmpty() ) emit lyricsNotFound( trackInformation );
    else  {

        // some special replacements for cleaning
        lyrics.replace( "{{", "(" );
        lyrics.replace( "}}", ")" );

        // construct lyrics information
        LyricsInformation lyricsInformation;
        lyricsInformation.setArtist( trackInformation.artist() );
        lyricsInformation.setTitle( trackInformation.titleOrFile() );
        lyricsInformation.setProvider( name() );
        lyricsInformation.setLyrics( lyrics );

        emit lyricsFound( trackInformation, lyricsInformation );

    }
}
