#ifndef MainWindow_h
#define MainWindow_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseMainWindow.h"
#include "Counter.h"
#include "CollectionFindWidget.h"
#include "CollectionItem.h"
#include "MpdCommand.h"
#include "MpdDefines.h"
#include "ObjectDeleter.h"
#include "StoredPlaylistItem.h"
#include "TimeStamp.h"
#include "TrackInformation.h"

#include <QBasicTimer>
#include <QCloseEvent>
#include <QMouseEvent>
#include <QTimerEvent>

#include <memory>

class AudioDeviceWidget;
class CollectionWidget;
class CurrentTrackWidget;
class DeleteTracksDialog;
class LogDialog;
class LyricsWidget;
class PlaybackControlToolBar;
class PlaylistWidget;
class ReconnectionWidget;
class SidePanelToolBar;
class SidePanelWidget;
class StoredPlaylistWidget;
class VolumeControlMenu;

#if WITH_TAGLIB
class TagEditionDialog;
#endif

namespace Mpd
{
    class InformationDialog;
}

//* main window
class MainWindow: public BaseMainWindow, private Base::Counter<MainWindow>
{

    Q_OBJECT

    public:

    //* constructor
    explicit MainWindow( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* mpd state
    Mpd::State state() const
    { return state_; }

    //* playback control
    PlaybackControlToolBar& playbackControlToolBar() const
    { return *playbackControlToolBar_; }

    //* current track widget
    CurrentTrackWidget& currentTrackWidget() const;

    //* volumem control menu
    VolumeControlMenu& volumeControlMenu() const;

    //* playlist widget
    PlaylistWidget& playlistWidget() const
    { return *playlistWidget_; }

    //* reconnection widget
    ReconnectionWidget& reconnectionWidget() const
    { return *reconnectionWidget_; }

    //* stored playlists
    StoredPlaylistWidget& storedPlaylistWidget() const;

    //* collection
    CollectionWidget& collectionWidget() const;

    //* lyrics
    LyricsWidget& lyricsWidget() const;

    #if WITH_CDIO_PARANOIA
    //* audiodevice
    AudioDeviceWidget& audioDeviceWidget() const;
    #endif

    //* log dialog
    LogDialog& logDialog() const
    { return *logDialog_; }

    //@}

    //*@name actions
    //@{

    //* log action
    QAction& showLogAction() const
    { return *showLogAction_; }

    //* log action
    QAction& toggleLogAction() const
    { return *toggleLogAction_; }

    //* collapse
    QAction& collapseAction() const
    { return *collapseAction_; }

    //@}

    //*@name modifiers
    //@{

    //* modifiers
    bool eventFilter( QObject*, QEvent* ) override;

    //* create context menu (overloaded)
    QMenu* createPopupMenu() override;

    //@}

    //* status flags
    /** needed to trigger property update of the DBus interface */
    enum StatusFlag
    {
        Volume = 1<<0,
        CurrentTrack = 1<<1,
        NextTrack = 1<<2,
        PreviousTrack = 1<<3,
        PlaybackState = 1<<4,
        PlaybackPosition = 1<<5,
        PlaybackMode = 1<<6
    };

    Q_DECLARE_FLAGS(StatusFlags, StatusFlag)

    Q_SIGNALS:

    //* request command
    void commandRequested( Mpd::Command );

    //* request command
    void commandsRequested( Mpd::Command::List );

    //* new notification
    void newNotification( TrackInformation );

    //* request icon
    void requestCover( CollectionItem );

    //* emitted when ellapsed time is set for audioStream track
    void elapsedTimeChanged( quint32 );

    //* emitted when playback status has changed
    void statusChanged( MainWindow::StatusFlags );

    public Q_SLOTS:

    //* clear
    void clear();

    //* save mpd version
    void saveMpdVersion( QString value )
    { mpdVersion_ = value; }

    //* get mpd statistics
    void mpdInformation();

    //* update status
    void updateStatus();

    //* clear playlist
    void clearPlaylist();

    //* update playlist
    void updatePlaylist();

    //* request collection update
    void updateCollection();

    //* request collection update
    void updateCollection( QString );

    //* update collection
    void loadCollection();

    //* update stored playlists
    void loadStoredPlaylists();

    //* add from filesystem
    void addFromFileSystem( const File::List& );

    //* replace from filesystem
    void replaceFromFileSystem( const File::List& );

    //* pause playback
    void pausePlayback();

    //* stop playback
    void stopPlayback();

    //* process command reply
    void processCommandReply( const Mpd::Command& );

    //* update cover
    void updateCover( const TrackInformation&, const TrackInformation& );

    protected:

    //* close event
    void closeEvent( QCloseEvent* ) override;

    //* timer event
    void timerEvent( QTimerEvent* ) override;

    //* mouse double click event
    void mouseDoubleClickEvent( QMouseEvent* ) override;

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration();

    //* toggle navigation frame visibility
    void _toggleSidePanelWidget( bool );

    //* splitter moved
    void _splitterMoved();

    //* add files
    void _addFiles( const File::List&, int );

    //* move tracks
    void _moveTracks( int );

    //* add tracks from collection
    void _addFromCollection( int );

    //* add collection items
    void _addFromCollection( const CollectionItem::List& );

    //* add collection items
    void _replaceFromCollection( const CollectionItem::List& );

    //* delete tracks from collection
    void _deleteTracksFromCollection( const CollectionItem::List& );

    //* search collection
    void _searchCollection( const QString&, TrackInformation::Mask );

    //* list playlist tracks
    void _listPlaylistTracks( const StoredPlaylistItem& );

    //* add stored playlist items
    void _addFromStoredPlaylist( int );

    //* add stored playlist items
    void _addFromStoredPlaylist( const StoredPlaylistItem::List& );

    //* add stored playlist items
    void _replaceFromStoredPlaylist( const StoredPlaylistItem::List& );

    //* remove from stored playlists
    void _removeFromStoredPlaylist( const StoredPlaylistItem::List& );

    //* add tracks to stored playlist
    void _addToStoredPlaylist( const StoredPlaylistItem&, int );

    //* move tracks to stored playlist
    void _moveToStoredPlaylist( const StoredPlaylistItem&, int );

    //* rename stored playlist
    void _renameStoredPlaylist( const StoredPlaylistItem&, const QString& );

    //* export stored playlist
    void _exportStoredPlaylist( const StoredPlaylistItem& );

    #if WITH_CDIO_PARANOIA
    //* add tracks from audio device
    void _addFromAudioDevice( int );

    //* add tracks from audio device
    void _addFromAudioDevice( const TrackInformation::List& );

    //* replace from audio device
    void _replaceFromAudioDevice( const TrackInformation::List& );
    #endif

    //* remove tracks
    void _removeSelectedTracks();

    //* save playlist
    void _savePlaylist();

    //* start playback
    void _startPlayback();

    //* resume playback
    void _resumePlayback();

    //* previous song
    void _previousTrack();

    //* next song
    void _nextTrack();

    //* set volume
    void _setVolume(int);

    //* seek
    void _seek(qreal);

    //* repeat none
    void _repeatNone();

    //* repeat one
    void _repeatOne();

    //* repeat all
    void _repeatAll();

    //* random
    void _random(bool);

    //* list albums
    void _listAlbums( const CollectionItem& );

    //* list albums
    void _listTracks( const CollectionItem& );

    #if WITH_TAGLIB
    //* edit tags for selected files
    void _editTags( const CollectionItem::List& );

    //* edit tags for selected files
    void _editTags( const File::List& );
    #endif

    //* show log
    void _showLog();

    //* toggle log
    void _toggleLog( bool );

    //* initialize dbus
    void _initDBus();

    //* deinit dbus
    void _deinitDBus();

    //* collapse main window
    void _toggleCollapse();

    //* set collapsed
    void _setCollapsed( bool );

    private:

    //* install actions
    void _installActions();

    //* set mpd state
    /* also propagate to child widgets */
    void _setState( Mpd::State );

    //* window title
    void _updateWindowTitle();

    //* update playback control buttons
    void _updatePlaybackControlButtons();

    //* notification
    void _notifyCurrentTrack();

    //* notify playback mode
    void _notifyPlaybackMode();

    //*@name command processing
    //@{

    //* status
    void _processStatusCommand( const Mpd::Command& );

    //* statistics
    void _processMpdInformationCommand( const Mpd::Command& );

    //* clear
    void _processClearCommand( const Mpd::Command& );

    //* update playlist
    void _processPlaylistInfoCommand( const Mpd::Command& );

    //* update stored playlist
    void _processStoredPlaylistInfoCommand( const Mpd::Command& );

    //* update
    void _processCollectionUpdateCommand( const Mpd::Command& );

    //* load collection
    void _processLoadCollectionCommand( const Mpd::Command::List& );

    //* list playlists
    void _processLoadPlaylistsCommand( const Mpd::Command::List& );

    //* find tracks
    void _processFindTracksCommand( const Mpd::Command& );

    //* find tracks
    void _processFindTrackCommand( const Mpd::Command& );

    //* search collection
    void _processCollectionSearchCommand( const Mpd::Command& );

    //@}

    //* add items from collection at given row
    void _doAddFromCollection( const CollectionItem::List&, int );

    //* add items from stired playlist at given row
    void _doAddFromStoredPlaylist( const StoredPlaylistItem::List&, int );

    #if WITH_CDIO_PARANOIA
    //* add items from audio device at a give row
    void _doAddFromAudioDevice( const TrackInformation::List&, int );
    #endif

    //* update covers from list of collection items
    void _updateCovers( const CollectionItem::List& );

    //* control toobar
    PlaybackControlToolBar* playbackControlToolBar_ = nullptr;

    //* main widget
    QWidget* mainWidget_ = nullptr;

    //* side panel toolbar
    SidePanelToolBar* sidePanelToolBar_ = nullptr;

    //* side panel
    SidePanelWidget* sidePanelWidget_ = nullptr;

    //* playlist
    PlaylistWidget* playlistWidget_ = nullptr;

    //* reconnection widget
    ReconnectionWidget* reconnectionWidget_ = nullptr;

    #if WITH_TAGLIB
    //* tag edition dialog
    TagEditionDialog* tagEditionDialog_ = nullptr;
    #endif

    //* delete tracks dialog
    DeleteTracksDialog* deleteTracksDialog_ = nullptr;

    //* mpd statistics
    Mpd::InformationDialog* mpdInformationDialog_ = nullptr;

    //* log dialog
    LogDialog* logDialog_ = nullptr;

    //*@name actions
    //@{

    //* log action
    QAction* showLogAction_ = nullptr;

    //* log action
    QAction* toggleLogAction_ = nullptr;

    //* collapse
    QAction* collapseAction_ = nullptr;

    //@}

    //* mpd state
    Mpd::State state_ = Mpd::State::Stopped;

    //* mpd version
    QString mpdVersion_;

    //* true if collapsed
    bool collapsed_ = false;

    //* expanded size
    QSize expandedSize_;
    QSize collapsedSize_;

    //*@name search
    //@{

    //* true if there is a search in progress
    bool searchActive_ = false;

    //* search mask
    TrackInformation::Mask searchMask_ = TrackInformation::None;

    //* search string
    QString searchedString_;

    //@}

    //*@name dbus
    //@{

    #if WITH_DBUS
    bool dbusEnabled_ = false;
    std::unique_ptr<QObject, Base::ObjectDeleter> dbusCoreAdaptor_;
    std::unique_ptr<QObject, Base::ObjectDeleter> dbusPlayerAdaptor_;
    #endif

    //@}

    //* keep track of hash keys for which icons have been requested
    QSet<uint> collectionItemKeys_;

    //* resize timer
    QBasicTimer resizeTimer_;

    //* collection update timer
    QBasicTimer collectionUpdateTimer_;

    //* collection update timestamp
    /** it is used to measured how much time it takes to update the db */
    TimeStamp collectionUpdateTimeStamp_;

};

Q_DECLARE_OPERATORS_FOR_FLAGS( MainWindow::StatusFlags )

#endif
