/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdCommand.h"

#include "CppUtil.h"
#include "Debug.h"

#if QT_VERSION < 0x040800
#include "CppUtil.h"
#endif

#include <QRegExp>

namespace Mpd
{

    //___________________________________________________________________
    const Command::NameMap& Command::_commandNames()
    {
        static const auto names = Base::makeT<NameMap>( {
            { Type::Undefined, QObject::tr( "unknown command" ) },
            { Type::LoadCollection, QObject::tr( "load collection" ) },
            { Type::UpdateCollection, QObject::tr( "update collection" ) },
            { Type::SearchCollection, QObject::tr( "search collection" ) },
            { Type::FindTrack, QObject::tr( "find track" ) },
            { Type::FindTracks, QObject::tr( "find tracks" ) },
            { Type::Play, QObject::tr( "play" ) },
            { Type::Resume, QObject::tr( "resume" ) },
            { Type::Stop, QObject::tr( "stop" ) },
            { Type::Pause, QObject::tr( "pause" ) },
            { Type::NextTrack, QObject::tr( "next track" ) },
            { Type::PreviousTrack, QObject::tr( "previous track" ) },
            { Type::ChangeVolume, QObject::tr( "change volume" ) },
            { Type::Seek, QObject::tr( "seek" ) },
            { Type::ChangeRepeatMode, QObject::tr( "change repeat mode" ) },
            { Type::PlaylistInfo, QObject::tr( "playlist info" ) },
            { Type::ClearPlaylist, QObject::tr( "clear playlist" ) },
            { Type::SavePlaylist, QObject::tr( "save playlist" ) },
            { Type::AddTracks, QObject::tr( "add tracks" ) },
            { Type::RemoveTracks, QObject::tr( "remove tracks" ) },
            { Type::MoveTracks, QObject::tr( "move tracks" ) },
            { Type::LoadPlaylists, QObject::tr( "load saved playlists" ) },
            { Type::StoredPlaylistInfo, QObject::tr( "saved playlist info" ) },
            { Type::Status, QObject::tr( "server status" ) },
            { Type::Statistics, QObject::tr( "server statistics" ) },
            { Type::Kill, QObject::tr( "kill server" ) }
        });

        return names;
    }

    //___________________________________________________________________
    int& Command::_counter()
    {
        static int counter( 0 );
        return counter;
    }

    //______________________________________________________________
    Command::Command():
        Counter( "Mpd::Command" ),
        id_( _counter()++ )
    {}

    //______________________________________________________________
    Command::Command( Type type, const QStringList& commandList ):
        Counter( "Mpd::Command" ),
        id_( _counter()++ ),
        type_( type ),
        commandList_( commandList )
    {}

    #if QT_VERSION < 0x040800
    //___________________________________________________________________
    Command::Command( Type type, std::initializer_list<QString>&& commandList ):
        Command( type, Base::makeT<QStringList>( std::move( commandList ) ) )
    {}
    #endif

    //______________________________________________________________
    Command::Command( Type type, const QString& command ):
        Command( type, QStringList( command ) )
    {}

    //______________________________________________________________
    Command::List Command::subCommands() const
    {

        Debug::Throw( "Mpd::Command::subCommands.\n" );

        // split reply
        QStringList replyList;

        {
            QRegExp regExp("(\n|^)(list_OK|OK|ACK).*\n" );
            regExp.setMinimal(true);

            int lastPosition(0);
            for( int position = 0; (position = reply_.indexOf( regExp, position )) >= 0; position += regExp.matchedLength() )
            {
                const int end( position + regExp.matchedLength() );
                replyList.append( reply_.mid( lastPosition, end-lastPosition ) );
                lastPosition = end;
            }
        }

        Command::List commands;

        auto commandIter( commandList_.begin() );
        for( auto replyIter = replyList.begin(); replyIter != replyList.end(); ++replyIter, ++commandIter )
        {
            Command command( type_, *commandIter );
            command.setReply( *replyIter );
            command.setOptionList( optionList_ );
            const auto lastLine( replyIter->split( "\n", QString::SkipEmptyParts).back() );
            command.setSuccess( lastLine.startsWith( "OK" ) || lastLine.startsWith( "list_OK" ) );

            commands.append( command );
        }

        for( ; commandIter != commandList_.end(); ++commandIter )
        {
            Command command( type_, *commandIter );
            command.setSuccess( false );
            commands.append( command );
        }

        return commands;

    }

    //_____________________________________________________________
    QTextStream& operator << ( QTextStream& out, const Command& command )
    {
        out << "Type: " << Base::toIntegralType( command.type_ ) << " commands: " << command.commandList_.join( "\n" );
        if( !command.optionList_.isEmpty() ) out << endl << "options: " << command.optionList_.join( "\n" );
        return out;
    }

}
