/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdInterface.h"

#include "Command.h"
#include "Debug.h"
#include "MpdBaseSocket.h"
#include "MpdServerConfiguration.h"

#include <signal.h>
#include <QDir>
#include <QRegExp>

namespace Mpd
{

    //______________________________________________
    Interface::Interface( QObject* parent ):
        QObject( parent ),
        Counter( "Mpd::Interface" )
    {}

    //______________________________________________
    Interface::~Interface() = default;

    //______________________________________________
    bool Interface::connect()
    {
        Debug::Throw( "Interface::connect.\n" );

        socket_.reset( BaseSocket::create( this, serverConfiguration_ ) );
        connect( socket_.get(), SIGNAL(error()), SIGNAL(socketError()) );
        connect( socket_.get(), SIGNAL(disconnected()), SLOT(_socketDisconnected()) );

        // try connect immediately
        if( !socket_->connect( serverConfiguration_ ) )
        {
            emit socketError();
            return false;
        }

        // read version from socket
        const auto reply = _readFromSocket();
        QRegExp regExp( "MPD\\s+(\\d+\\.\\d+\\.\\d+)" );
        if( reply.indexOf( regExp ) >= 0 )
        {
            const auto version = regExp.cap(1);
            if( version != version_ )
            {
                version_ = version;
                emit versionAvailable( version_ );
            }
        }

        return true;
    }

    //______________________________________________
    void Interface::disconnect()
    {
        Debug::Throw( "Mpd::Interface::disconnect.\n" );

        if( socket_ )
        {
            socket_->disconnect();
            socket_.reset();
        }

    }

    //______________________________________________
    bool Interface::addCommand( Command command )
    {

        // connect socket
        if( !( _isSocketConnected() || connect() ) )
        {
            emit commandFailed( command );
            return false;
        }

        const auto commandList = command.commandList();
        if( commandList.empty() )
        {

            command.setReply( "OK\n" );
            emit commandCompleted( command );
            return true;

        } else if( commandList.size() == 1 ) {

            const auto commandString( commandList.front() );
            _sendCommand( commandString );

        } else {

            _sendCommand( "command_list_ok_begin" );
            for( const auto& commandString:commandList ) { _sendCommand( commandString ); }
            _sendCommand( "command_list_end" );

        }

        auto reply = _readFromSocket();
        Debug::Throw() << "Mpd::Interface::addCommand -"
            << command
            << " Reply: " << reply
            << endl;

        if( reply.isEmpty() )
        {

            emit commandFailed( command );
            return false;

        } else {

            command.setReply( reply );
            const auto replyList = reply.split( '\n', QString::SkipEmptyParts );
            if( !replyList.isEmpty() && replyList.back().startsWith( "OK" ) ) command.setSuccess( true );

            emit commandCompleted( command );
            return true;

        }

    }

    //______________________________________________
    bool Interface::addCommands( Command::List commands )
    {
        for( const auto& command: commands )
        { if( !addCommand( command ) ) return false; }

        return true;
    }

    //______________________________________________
    void Interface::_socketDisconnected()
    {
        Debug::Throw( "Interface::_socketDisconnected.\n" );
        emit socketError();
    }

    //______________________________________________
    bool Interface::_isSocketConnected() const
    { return socket_ && socket_->isConnected(); }

    //______________________________________________
    void Interface::_sendCommand( QString command )
    {
        if( !command.endsWith( '\n' ) ) command += '\n';
        socket_->write( command.toUtf8() );
    }

    //______________________________________________
    QString Interface::_readFromSocket()
    {
        QRegExp regExp( "(\n|^)(OK|ACK).*\n" );
        regExp.setMinimal(true);

        QString reply;
        while( _isSocketConnected() )
        {
            if( !socket_->waitForReadyRead() )
            {
                socket_->close();
                return QString();
            }

            reply.append( QString::fromUtf8( socket_->readAll() ) );
            const auto position = reply.indexOf( regExp );
            if( position >= 0 )
            {
                reply = reply.left( position + regExp.matchedLength() );
                return reply;
            }
        }

        return QString();
    }

}
