#ifndef PlaylistModel_h
#define PlaylistModel_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "CoverCache.h"
#include "Debug.h"
#include "ListModel.h"
#include "MpdDefines.h"
#include "TrackInformationModel.h"

#include <QIcon>
#include <QMimeData>


class PlaylistModel: public TrackInformationModel
{

    Q_OBJECT

    public:

    //* constructor
    explicit PlaylistModel( QObject* = nullptr );

    //*@name accessors
    //@{

    //* flags
    Qt::ItemFlags flags( const QModelIndex& ) const override;

    //* mime type
    QStringList mimeTypes() const override;

    //* mime data
    QMimeData* mimeData( const QModelIndexList& ) const override;

    //* current track index
    QModelIndex currentTrackIndex() const override
    { return currentTrackIndex_; }

    //* current track
    const TrackInformation& currentTrack() const
    { return get( currentTrackIndex_ ); }

    //* true if previous track is valid
    bool hasPreviousTrack() const
    { return contains( _previousTrackIndex() ); }

    //* true if next track is valid
    bool hasNextTrack() const
    { return contains( _nextTrackIndex() ); }

    //* state
    Mpd::State state() const
    { return state_; }

    //* state icon
    QIcon stateIcon() const override
    { return stateIcon_; }

    //@}

    //*@name modifiers
    //@{

    //* state
    void setState( Mpd::State );

    //* repeat mode
    void setRepeatMode( Mpd::RepeatMode value )
    { repeatMode_ = value; }

    //* drop mine data
    bool dropMimeData( const QMimeData*, Qt::DropAction, int, int, const QModelIndex& ) override;

    //* clear
    void clear() override
    {
        ListModel::clear();
        currentTrackIndex_ = QModelIndex();
    }

    //@}

    Q_SIGNALS:

    //* request adding files
    void requestAddFiles( File::List, int );

    //* request adding tracks from collection
    void requestAddFromCollection( int );

    //* request add tracks from stored playlist
    void requestAddFromStoredPlaylist( int );

    #if WITH_CDIO_PARANOIA
    //* request adding tracks from audio device
    void requestAddFromAudioDevice( int );
    #endif

    //* request moving tracks
    void requestMoveTracks( int );

    public Q_SLOTS:

    //* set current track index
    bool setCurrentTrackIndex( const QModelIndex& );

    //* select previous track
    bool selectPreviousTrack();

    //* select next track
    bool selectNextTrack();

    private:

    //* mpd state
    void _setState( Mpd::State );

    //* previous track index
    QModelIndex _previousTrackIndex() const;

    //* next track index
    QModelIndex _nextTrackIndex() const;

    //* current track
    QModelIndex currentTrackIndex_;

    //* state
    Mpd::State state_ = Mpd::State::Stopped;

    //* repeat mode
    Mpd::RepeatMode repeatMode_ = Mpd::RepeatMode::RepeatNone;

    //* state icon
    QIcon stateIcon_;

};

#endif
