/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistParser.h"

#include "CppUtil.h"
#include "Debug.h"
#include "XmlError.h"

#include <QDomDocument>
#include <QFile>
#include <QUrl>

//_____________________________________________________________________
bool PlaylistParser::isPlaylist( const File& file )
{
    static const auto extensions( Base::makeT<QStringList>( { "m3u", "xspf" } ) );
    return extensions.contains( file.extension() );
}

//_____________________________________________________________________
PlaylistParser::PlaylistParser( const File& file ):
    Counter( "PlaylistParser" ),
    file_( file )
{ Debug::Throw() << "PlaylistParser::PlaylistParser - file: " << file << endl; }

//_____________________________________________________________________
TrackInformation::List PlaylistParser::parse() const
{
    Debug::Throw( "PlaylistParser::parse.\n" );
    if( !file_.exists() ) return TrackInformation::List();
    const auto extension( file_.extension().get() );
    if( extension == QLatin1String( "m3u" ) ) return _parseM3u();
    else if( extension == QLatin1String( "xspf" ) ) return _parseXspf();
    else return TrackInformation::List();
}

//_____________________________________________________________________
TrackInformation::List PlaylistParser::_parseM3u() const
{
    Debug::Throw( "PlaylistParser::_parseM3u.\n" );

    QFile qfile( file_ );
    if( !qfile.open( QIODevice::ReadOnly ) ) return TrackInformation::List();

    const auto path( file_.path() );
    TrackInformation::List trackInformationList;
    QString line;
    while( !( line = qfile.readLine() ).isNull() )
    {

        if( line.startsWith( '#' ) ) continue;

        File file( line.trimmed() );
        if( !file.isAbsolute() ) file.addPath( path );
        if( file.exists() )
        {

            TrackInformation trackInformation;
            trackInformation.setFile( file );
            trackInformationList.append( trackInformation );

        }

    }

    return trackInformationList;
}

//_____________________________________________________________________
TrackInformation::List PlaylistParser::_parseXspf() const
{
    Debug::Throw( "PlaylistParser::_parseXspf.\n" );
    QFile qfile( file_ );
    if( !qfile.open( QIODevice::ReadOnly ) ) return TrackInformation::List();

    // create document and set content
    QDomDocument document;
    XmlError error;
    if( !document.setContent( &qfile, error ) ) return TrackInformation::List();

    TrackInformation::List trackInformationList;
    const auto nodeList( document.elementsByTagName( QLatin1String( "track" ) ) );
    for( int i = 0; i < nodeList.count(); ++i )
    {
        TrackInformation trackInformation;

        // get element
        const auto element( nodeList.at( i ).toElement() );

        // parse children elements
        for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
        {
            const auto childElement = childNode.toElement();
            if( childElement.isNull() ) continue;

            const auto tagName( childElement.tagName() );
            if( tagName == QLatin1String( "location" ) )
            {
                const auto path( QUrl::fromEncoded( childElement.text().toUtf8() ).path() );
                trackInformation.setFile( File( path ) );
            } else if( tagName == QLatin1String( "title" ) ) trackInformation.setTitle( childElement.text() );
            else if( tagName == QLatin1String( "creator" ) ) trackInformation.setArtist( childElement.text() );
            else if( tagName == QLatin1String( "album" ) ) trackInformation.setAlbum( childElement.text() );
            else if( tagName == QLatin1String( "trackNum" ) ) trackInformation.setTrack( childElement.text().toInt() );
            else if( tagName == QLatin1String( "duration" ) ) trackInformation.setDuration( childElement.text().toInt()/1000 );
        }

        trackInformationList.append( trackInformation );
    }

    return trackInformationList;

}
