/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SidePanelWidget.h"

#include "BaseStatusBar.h"
#include "CollectionWidget.h"
#include "Debug.h"
#include "FileSystemWidget.h"
#include "IconSize.h"
#include "LyricsWidget.h"
#include "Singleton.h"
#include "StoredPlaylistWidget.h"
#include "Util.h"
#include "XmlOptions.h"

#if WITH_CDIO_PARANOIA
#include "AudioDeviceWidget.h"
#endif

#include <QPainter>

#include <QHeaderView>
#include <QLayout>

//_______________________________________________________________
SidePanelWidget::SidePanelWidget( QWidget* parent ):
    QWidget( parent )
{

    Debug::Throw( "SidePanelWidget:SidePanelWidget.\n" );
    auto vLayout = new QVBoxLayout();
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( stackedWidget_ = new QStackedWidget( this ) );

    connect( stackedWidget_, SIGNAL(currentChanged(int)), this, SIGNAL(currentChanged(int)) );

    // stack widget
    stackedWidget_->addWidget( collectionWidget_ = new CollectionWidget );
    stackedWidget_->addWidget( storedPlaylistWidget_ = new StoredPlaylistWidget );
    stackedWidget_->addWidget( lyricsWidget_ = new LyricsWidget );
    stackedWidget_->addWidget( fileSystemWidget_ = new FileSystemWidget );

    #if WITH_CDIO_PARANOIA
    stackedWidget_->addWidget( audioDeviceWidget_ = new AudioDeviceWidget );
    #endif

    // statusbar
    vLayout->addWidget( statusBar_ = new BaseStatusBar( this ) );
    statusBar_->addLabel(1);
    statusBar_->label(0).setMinimumHeight(
        IconSize::Small
        + 2*statusBar_->label(0).margin() + 1 );

    // customize fileSystem Widget
    fileSystemWidget_->workingDirectoryAction().setVisible( false );

    // current widget
    setCurrentWidget( collectionWidget_ );

    // connections
    connect( this, SIGNAL(currentChanged(int)), SLOT(_updateCurrentWidget()) );

    // actions
    _installActions();

    // configuration
    auto application( Base::Singleton::get().application() );
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

    connect( application, SIGNAL(saveConfiguration()), SLOT(_saveConfiguration()) );

}

//______________________________________________________________________
void SidePanelWidget::_updateCurrentWidget()
{
    Debug::Throw( "SidePanelWidget::_updateCurrentWidget.\n" );
    if( stackedWidget_->currentWidget() ) stackedWidget_->currentWidget()->update();
    if( stackedWidget_->currentWidget() == lyricsWidget_ ) lyricsWidget_->requestLyrics();
    #if WITH_CDIO_PARANOIA
    if( stackedWidget_->currentWidget() == audioDeviceWidget_ ) audioDeviceWidget_->tryReadDevice();
    #endif
}

//______________________________________________________________________
void SidePanelWidget::_updateConfiguration()
{
    Debug::Throw( "SidePanelWidget::_updateConfiguration.\n" );

    //* working path
    File workingPath( XmlOptions::get().raw( "MUSIC_DIR" ) );

    //* current path
    File currentPath;
    if( !( XmlOptions::get().contains( "FILESYSTEM_CURRENT_PATH" ) &&
        (currentPath = File( XmlOptions::get().raw( "FILESYSTEM_CURRENT_PATH" ) ) ).exists() ) )
    { currentPath = workingPath; }

    // assign
    if( currentPath.exists() ) fileSystemWidget_->setPath( currentPath );
    if( workingPath.exists() ) fileSystemWidget_->setHomePath( workingPath );
}


//______________________________________________________________________
void SidePanelWidget::_saveConfiguration()
{

    Debug::Throw( "SidePanelWidget::_saveConfiguration.\n" );
    if( !fileSystemWidget_ ) return;

    const auto currentPath( fileSystemWidget_->path() );
    if( currentPath.exists() )
    { XmlOptions::get().set( "FILESYSTEM_CURRENT_PATH", currentPath ); }

}

//______________________________________________________________________
void SidePanelWidget::_installActions()
{

    Debug::Throw( "SidePanelWidget::_installActions.\n" );
    addAction( visibilityAction_ = new QAction( tr( "Show &Navigation Panel" ), this ) );
    visibilityAction_->setCheckable( true );
    visibilityAction_->setChecked( true );
    connect( visibilityAction_, SIGNAL(toggled(bool)), SLOT(setVisible(bool)) );

}
