/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "StoredPlaylistModel.h"

#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"

#include <QMimeData>

//_______________________________________________________________________________________
StoredPlaylistModel::StoredPlaylistModel( QObject* parent ):
    TreeModel( parent ),
    Counter( "StoredPlaylistModel" )
{ Debug::Throw( "StoredPlaylistModel::StoredPlaylistModel.\n" ); }

//_______________________________________________________________________________________
Qt::ItemFlags StoredPlaylistModel::flags( const QModelIndex& index ) const
{
    if( index.isValid() )
    {
        const auto& item( get( index ) );
        if( item.type() == StoredPlaylistItem::Type::Playlist )
        {

            return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;

        } else return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled;

    } else return Qt::ItemIsDropEnabled;

}

//__________________________________________________________________
QStringList StoredPlaylistModel::mimeTypes() const
{
    return Base::makeT<QStringList>({ TrackInformation::MimeType, StoredPlaylistItem::MimeType });
}

//______________________________________________________________________
QMimeData* StoredPlaylistModel::mimeData(const QModelIndexList &indexes) const
{
    if( indexes.empty() ) return nullptr;
    auto mime = new QMimeData;
    mime->setData( StoredPlaylistItem::MimeType, nullptr );
    return mime;
}

//__________________________________________________________________
QVariant StoredPlaylistModel::data( const QModelIndex& index, int role ) const
{

    if( !index.isValid() ) return QVariant();

    switch( role )
    {
        case Qt::DisplayRole:
        {

            const auto& item( _find( index.internalId() ).get() );
            switch( item.type() )
            {
                case StoredPlaylistItem::Type::Playlist:
                return item.displayName();

                case StoredPlaylistItem::Type::Track:
                return item.titleOrFile();

                default: return QVariant();

            }
        }

        case Qt::DecorationRole:
        {

            const auto& item( _find( index.internalId() ).get() );
            switch( item.type() )
            {
                case StoredPlaylistItem::Type::Playlist: return IconEngine::get( IconNames::Playlist );
                case StoredPlaylistItem::Type::Track: return IconEngine::get( IconNames::Track );
                default: return QVariant();
            }

        }

        default: return QVariant();
    }

}

//__________________________________________________________________
QVariant StoredPlaylistModel::headerData(int section, Qt::Orientation orientation, int role) const
{

    switch( role )
    {
        case Qt::DisplayRole:
        {
            switch( section )
            {
                case Name: return tr( "Name" );
                default: return QVariant();
            }
            break;
        }

        default: return QVariant();
    }

    // return empty
    return QVariant();

}

//__________________________________________________________________
bool StoredPlaylistModel::hasChildren( const QModelIndex& parent ) const
{
    if( !parent.isValid() ) return TreeModel::hasChildren( parent );

    const auto item( get( parent ) );
    return item.type() == StoredPlaylistItem::Type::Playlist;
}

//__________________________________________________________________
QIcon StoredPlaylistModel::icon( const StoredPlaylistItem& item ) const
{ return cache_ ? cache_->icon( CollectionItem( item ) ):QIcon(); }

//__________________________________________________________________
bool StoredPlaylistModel::dropMimeData(const QMimeData* data , Qt::DropAction action, int row, int, const QModelIndex& parent )
{

    // drop from current playlist
    if( data->hasFormat( TrackInformation::MimeType ) )
    {

        if( !parent.isValid() ) return true;

        const auto parentItem( get( parent ) );
        if( parentItem.type() == StoredPlaylistItem::Type::Playlist )
        { emit requestAddToPlaylist( parentItem, row ); }

    } else if( data->hasFormat( StoredPlaylistItem::MimeType ) ) {

        if( !parent.isValid() ) return true;

        const auto parentItem( get( parent ) );
        if( parentItem.type() == StoredPlaylistItem::Type::Playlist )
        { emit requestMoveToPlaylist( parentItem, row ); }

    }

    return true;
}

//__________________________________________________________________
bool StoredPlaylistModel::updatePlaylist( const StoredPlaylistItem& other )
{
    if( other.type() != StoredPlaylistItem::Type::Playlist ) return false;
    auto item( _root().find( other ) );
    if( item )
    {

        item->get().setTrackCount( other.trackCount() );
        item->get().setDuration( other.duration() );
        return true;

    } else return false;

}

//__________________________________________________________________
bool StoredPlaylistModel::updatePlaylist( const StoredPlaylistItem& other, const QString& playlist )
{
    if( other.type() != StoredPlaylistItem::Type::Playlist ) return false;
    auto item( _root().find( other ) );
    if( item )
    {

        item->get().setPlaylist( playlist );
        return true;

    } else return false;

}

//__________________________________________________________________
void StoredPlaylistModel::setCoverCache( const CoverCache::Pointer& value )
{
    cache_ = value;
    connect( cache_.get(), SIGNAL(modified()), SIGNAL(layoutAboutToBeChanged()) );
    connect( cache_.get(), SIGNAL(modified()), SIGNAL(layoutChanged()) );
}
