/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "StoredPlaylistWidget.h"

#include "Debug.h"
#include "ContextMenu.h"
#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "ListNotificationWidget.h"
#include "SavePlaylistDialog.h"
#include "ScrollBarMonitor.h"
#include "StoredPlaylistItemToolTipWidget.h"
#include "StoredPlaylistView.h"
#include "WarningDialog.h"
#include "XmlOptions.h"

#include <QLayout>

#include <algorithm>
#include <numeric>

//________________________________________________________
StoredPlaylistWidget::StoredPlaylistWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "StoredPlaylistWidget" )
{
    Debug::Throw( "StoredPlaylistWidget::StoredPlaylistWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( list_ = new StoredPlaylistView( this ) );
    list_->setModel( &model_ );
    list_->setOptionName( "STORED_PLAYLISTS" );

    scrollBarMonitor_ = new ScrollBarMonitor( list_ );

    // notification widget
    notificationWidget_ = new ListNotificationWidget(  tr( "No playlist found" ), list_ );
    notificationWidget_->show();

    // tooltip widget
    toolTipWidget_ = new StoredPlaylistItemToolTipWidget( this );

    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_itemActivated(QModelIndex)) );
    connect( list_, SIGNAL(expanded(QModelIndex)), SLOT(_itemExpanded(QModelIndex)) );
    connect( list_, SIGNAL(hovered(QModelIndex)), SLOT(_showToolTip(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );

    {
        // context menu
        auto menu = new ContextMenu( list_ );
        menu->addAction( addToPlaylistAction_ );
        menu->addAction( replacePlaylistAction_ );
        menu->addSeparator();
        menu->addAction( deleteItemAction_ );
        menu->addAction( renamePlaylistAction_ );
        menu->addAction( exportPlaylistAction_ );
        menu->addSeparator();
        menu->addAction( &list_->collapseAllAction() );
        menu->setHideDisabledActions( true );

        connect( menu, SIGNAL(aboutToShow()), SLOT(_updateActions()) );
    }

}

//________________________________________________________
StoredPlaylistItem::List StoredPlaylistWidget::expandedItems() const
{
    // store model expanded items
    StoredPlaylistItem::List out;
    auto indexes = model_.indexes();

    // get items for all indexes that are expanded
    std::transform(
        indexes.begin(),
        std::remove_if( indexes.begin(), indexes.end(),
        [this]( const QModelIndex& index ) { return !list_->isExpanded( index ); } ),
        std::back_inserter( out ),
        [this]( const QModelIndex& index ) { return model_.get( index ); } );

    return out;
}

//________________________________________________________
void StoredPlaylistWidget::updatePlaylists( const StoredPlaylistItem::List& items )
{
    Debug::Throw( "StoredPlaylistWidget::addTracks.\n" );
    notificationWidget_->setVisible( items.isEmpty() );

    // split items by album and assign to model
    for( auto iter = items.begin(); iter != items.end(); )
    {
        StoredPlaylistItem playlist( *iter );
        playlist.setType( StoredPlaylistItem::Type::Playlist );
        StoredPlaylistItem::List copy;
        StoredPlaylistItem::List::const_iterator next;
        std::copy(
            iter,
            next = std::find_if_not( iter, items.end(), StoredPlaylistItem::SamePlaylistFTor(playlist) ),
            std::back_inserter( copy ) );

        // calculate total duration and assign to album
        playlist.setDuration( std::accumulate( copy.begin(), copy.end(), qint32(0),
            [](qint32 duration, const StoredPlaylistItem& item )
            { return duration + item.duration(); } ) );

        playlist.setTrackCount( copy.size() );

        // replace playlist in model
        model_.updatePlaylist( playlist );

        // assign songs
        model_.set( playlist, copy );
        iter = next;
    }
}

//________________________________________________________
void StoredPlaylistWidget::saveExpandedItems()
{
    Debug::Throw( "StoredPlaylistWidget::saveExpandedItems.\n" );
    expandedItems_ = expandedItems();
}

//________________________________________________________
void StoredPlaylistWidget::restoreExpandedItems()
{
    Debug::Throw( "StoredPlaylistWidget::restoreExpandedItems.\n" );
    for( const auto& item:expandedItems_ )
    {
        const auto index( model_.index( item ) );
        if( index.isValid() ) list_->setExpanded( index, true );
    }

    expandedItems_.clear();
}

//________________________________________________________
void StoredPlaylistWidget::lock()
{
    locked_ = true;
    list_->setUpdatesEnabled( false );
    scrollBarMonitor_->save();
}

//________________________________________________________
void StoredPlaylistWidget::unlock()
{
    scrollBarMonitor_->restore();
    list_->setUpdatesEnabled( true );
    locked_ = false;
}

//________________________________________________________
void StoredPlaylistWidget::_itemActivated( const QModelIndex& index )
{
    Debug::Throw()
        << "StoredPlaylistWidget::_itemActivated -"
        << " index: " << index.row() << " " << index.column()
        << endl;

    // find item
    const auto& item( model_.get(index) );

    // emit signal
    switch( item.type() )
    {
        default: break;

        case StoredPlaylistItem::Type::Playlist:
        list_->setExpanded( index, !list_->isExpanded( index ) );
        break;

        case StoredPlaylistItem::Type::Track:
        emit requestAddToPlaylist( Base::makeT<StoredPlaylistItem::List>({item}) );
        break;

    }

}

//________________________________________________________
void StoredPlaylistWidget::_itemExpanded( const QModelIndex& index )
{
    Debug::Throw()
        << "StoredPlaylistWidget::_itemExpanded -"
        << " index: " << index.row() << " " << index.column()
        << " locked: " << locked_
        << endl;

    // do nothing if locked
    if( locked_ ) return;

    // find item
    const auto& item( model_.get(index) );

    // emit signal
    switch( item.type() )
    {
        default: break;
        case StoredPlaylistItem::Type::Playlist: emit requestTracks( item ); break;
    }
}

//________________________________________________________
void StoredPlaylistWidget::_updateActions()
{
    const auto selectedItems( model_.get( list_->selectionModel()->selectedRows() ) );
    const bool hasSelection( !selectedItems.empty() );
    addToPlaylistAction_->setEnabled( hasSelection );
    replacePlaylistAction_->setEnabled( hasSelection );

    const auto selectedPlaylists( std::count_if( selectedItems.begin(), selectedItems.end(), StoredPlaylistItem::SameTypeFTor( StoredPlaylistItem::Type::Playlist ) ) );
    const auto selectedTracks( std::count_if( selectedItems.begin(), selectedItems.end(), StoredPlaylistItem::SameTypeFTor( StoredPlaylistItem::Type::Track ) ) );

    if( selectedPlaylists > 0 && selectedTracks == 0 )
    {

        deleteItemAction_->setEnabled( true );
        deleteItemAction_->setText( selectedPlaylists > 1 ? tr( "Delete Playlists" ):tr( "Delete Playlist" ) );

    } else if( selectedTracks > 0 && selectedPlaylists == 0 ) {

        deleteItemAction_->setEnabled( true );
        deleteItemAction_->setText( tr( "Remove From Playlist" ) );

    } else deleteItemAction_->setEnabled( false );

    renamePlaylistAction_->setEnabled( selectedPlaylists == 1 && selectedTracks == 0 );
    exportPlaylistAction_->setEnabled( selectedPlaylists == 1 && selectedTracks == 0 );

}

//________________________________________________________
void StoredPlaylistWidget::_addSelectionToPlaylist()
{
    Debug::Throw( "StoredPlaylistWidget::_addSelectionToPlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestAddToPlaylist( items );
}

//________________________________________________________
void StoredPlaylistWidget::_replacePlaylist()
{
    Debug::Throw( "StoredPlaylistWidget::_replacePlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestReplacePlaylist( items );
}

//________________________________________________________
void StoredPlaylistWidget::_deleteItems()
{
    Debug::Throw( "StoredPlaylistWidget::_deleteItems.\n" );
    const auto items( model_.get( list_->selectionModel()->selectedRows() ) );

    // remove playlists from model directly
    auto copy( items );
    copy.erase( std::remove_if( copy.begin(), copy.end(), StoredPlaylistItem::DifferentTypeFTor( StoredPlaylistItem::Type::Playlist ) ), copy.end() );
    model_.remove( copy );

    if( !items.empty() ) emit requestDeleteItems( items );
}

//________________________________________________________
void StoredPlaylistWidget::_renamePlaylist()
{
    Debug::Throw( "StoredPlaylistWidget::_renamePlaylist.\n" );

    // check selection
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( items.size() != 1 ) return;

    // check item
    const auto item( items.front() );
    if( item.type() != StoredPlaylistItem::Type::Playlist ) return;

    // dialog
    SavePlaylistDialog dialog( this );
    dialog.setPlaylist( item.displayName() );
    dialog.setWindowTitle( tr( "Rename Playlist" ) );
    if( !dialog.exec() ) return;

    // check new name
    const auto playlist( dialog.playlist() );
    if( playlist.isEmpty() || item.playlist() == playlist ) return;

    // update model
    model_.updatePlaylist( item, playlist );

    // emit request
    emit requestRenamePlaylist( item, playlist );
}

//________________________________________________________
void StoredPlaylistWidget::_exportPlaylist()
{
    Debug::Throw( "StoredPlaylistWidget::_exportPlaylist.\n" );

    // check item
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( items.size() != 1 ) return;

    // check item
    const auto item( items.front() );
    if( item.type() != StoredPlaylistItem::Type::Playlist ) return;

    emit requestExportPlaylist(item);
}

//________________________________________________________
void StoredPlaylistWidget::_showToolTip( const QModelIndex& index )
{
    Debug::Throw( "StoredPlaylistWidget::_showToolTip.\n" );

    // check index
    if( !index.isValid() ) toolTipWidget_->hide();
    else {

        // get collection item and check
        const auto& item( model_.get( index ) );

        // get icon
        auto icon( model_.icon( item ) );
        if( icon.isNull() ) icon = IconEngine::get(
            item.type() == StoredPlaylistItem::Type::Playlist ?
            IconNames::Playlist:
            IconNames::Track );

        // assign item
        static_cast<StoredPlaylistItemToolTipWidget*>( toolTipWidget_ )->setStoredPlaylistItem( item, icon );

        // move tooltip
        toolTipWidget_->setIndexRect(
            list_->visualRect( index ).
            translated( list_->viewport()->mapToGlobal( QPoint( 0, 0 ) ) ) );

        // show
        toolTipWidget_->showDelayed();

    }

}

//________________________________________________________
void StoredPlaylistWidget::_installActions()
{
    Debug::Throw( "StoredPlaylistWidget::_installActions.\n" );

    addAction( addToPlaylistAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add to Playlist" ), this ) );
    connect( addToPlaylistAction_, SIGNAL(triggered()), SLOT(_addSelectionToPlaylist()) );
    addToPlaylistAction_->setEnabled( false );

    addAction( replacePlaylistAction_ = new QAction( tr( "Replace Playlist" ), this ) );
    connect( replacePlaylistAction_, SIGNAL(triggered()), SLOT(_replacePlaylist()) );
    replacePlaylistAction_->setEnabled( false );

    addAction( renamePlaylistAction_ = new QAction( tr( "Rename Playlist" ), this ) );
    renamePlaylistAction_->setShortcut( Qt::Key_F2 );
    renamePlaylistAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( renamePlaylistAction_, SIGNAL(triggered()), SLOT(_renamePlaylist()) );

    addAction( deleteItemAction_ = new QAction( IconEngine::get( IconNames::Remove ), tr( "Delete Selected Tracks" ), this ) );
    deleteItemAction_->setShortcut( QKeySequence::Delete );
    deleteItemAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( deleteItemAction_, SIGNAL(triggered()), SLOT(_deleteItems()) );
    deleteItemAction_->setEnabled( false );

    addAction( exportPlaylistAction_ = new QAction( IconEngine::get( IconNames::Export ), tr( "Export Playlist" ), this ) );
    connect( exportPlaylistAction_, SIGNAL(triggered()), SLOT(_exportPlaylist()) );
    exportPlaylistAction_->setEnabled( false );


}
