/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TagEditionDialog.h"

#include "CppUtil.h"
#include "CustomComboBox.h"
#include "Debug.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "LineEditor.h"
#include "TagLibInterface.h"

#include <QLabel>

#include <algorithm>

namespace
{

    namespace Local
    {
        class SpinBox: public QSpinBox, private Base::Counter<SpinBox>
        {

            Q_OBJECT

                public:

                //* constructor
                SpinBox( QWidget* parent = nullptr ):
                QSpinBox( parent ),
                Counter( "SpinBox" )
            {}

            protected:

            //* convert value from text
            QString textFromValue( int value ) const override
            {
                if( value == minimum() ) return QString();
                else return QSpinBox::textFromValue( value );
            }

        };

    }

}

//____________________________________________________________
TagEditionDialog::TagEditionDialog( QWidget* parent ):
    CustomDialog( parent, OkButton|CancelButton )
{

    Debug::Throw( "TagEditionDialog::TagEditionDialog.\n" );
    setWindowTitle( "Track Information" );
    setOptionName( "TAG_EDITION_DIALOG" );

    auto gridLayout = new GridLayout;
    gridLayout->setMaxCount(2);
    gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
    mainLayout().addLayout( gridLayout );

    QLabel* label;

    // location
    gridLayout->addWidget( label = new QLabel( tr( "Location:" ), this ) );
    gridLayout->addWidget( locationEditor_ = new LineEditor( this ) );
    label->setBuddy( locationEditor_ );
    locationEditor_->setReadOnly(true);

    // artist
    gridLayout->addWidget( label = new QLabel( tr( "Artist:" ), this ) );
    gridLayout->addWidget( artistEditor_ = new CustomComboBox( this ) );
    artistEditor_->setEditable(true);
    artistEditor_->setAutoCompletion(true, Qt::CaseSensitive);
    artistEditor_->setSizeAdjustPolicy( QComboBox::AdjustToMinimumContentsLengthWithIcon );
    label->setBuddy( artistEditor_ );
    connect( artistEditor_->lineEdit(), SIGNAL(textChanged(QString)), SLOT(_updateButtons()) );

    // album
    gridLayout->addWidget( label = new QLabel( tr( "Album:" ), this ) );
    gridLayout->addWidget( albumEditor_ = new CustomComboBox( this ) );
    albumEditor_->setEditable(true);
    albumEditor_->setAutoCompletion(true, Qt::CaseSensitive);
    albumEditor_->setSizeAdjustPolicy( QComboBox::AdjustToMinimumContentsLengthWithIcon );
    label->setBuddy( albumEditor_ );
    connect( albumEditor_->lineEdit(), SIGNAL(textChanged(QString)), SLOT(_updateButtons()) );

    // title
    gridLayout->addWidget( label = new QLabel( tr( "Track title:" ), this ) );
    gridLayout->addWidget( titleEditor_ = new LineEditor( this ) );
    label->setBuddy( titleEditor_ );
    connect( titleEditor_, SIGNAL(textEdited(QString)), SLOT(_updateButtons()) );

    // year
    gridLayout->addWidget( label = new QLabel( tr( "Year:" ), this ) );
    gridLayout->addWidget( yearEditor_ = new Local::SpinBox( this ) );
    label->setBuddy( yearEditor_ );
    yearEditor_->setMaximum( 3000 );
    connect( yearEditor_, SIGNAL(valueChanged(int)), SLOT(_updateButtons()) );

    // disc
    gridLayout->addWidget( label = new QLabel( tr( "Disc number:" ), this ) );
    gridLayout->addWidget( discNumberEditor_ = new Local::SpinBox( this ) );
    label->setBuddy( discNumberEditor_ );
    discNumberEditor_->setMaximum( 10 );
    connect( discNumberEditor_, SIGNAL(valueChanged(int)), SLOT(_updateButtons()) );

    // track number
    gridLayout->addWidget( label = new QLabel( tr( "Track number:" ), this ) );
    gridLayout->addWidget( trackNumberEditor_ = new Local::SpinBox( this ) );
    label->setBuddy( trackNumberEditor_ );
    trackNumberEditor_->setMaximum( 100 );
    connect( trackNumberEditor_, SIGNAL(valueChanged(int)), SLOT(_updateButtons()) );

    // add label about writable files
    gridLayout->addWidget( readOnlyLabel_ = new QLabel( tr( "Warning: selected files are not writable. Modifications will not be saved." ), this ), gridLayout->currentRow(), 0, 1, 2 );
    readOnlyLabel_->setAlignment( Qt::AlignCenter );
    readOnlyLabel_->setMargin( 10 );
    readOnlyLabel_->hide();

    // customize buttons
    okButton().setText( tr( "Save and Close" ) );
    okButton().setIcon( IconEngine::get( IconNames::Save ) );
    connect( &okButton(), SIGNAL(clicked()), SLOT(_saveTags()) );
    okButton().setEnabled( false );

}

//____________________________________________________________
File TagEditionDialog::location() const
{ return File( locationEditor_->text() ); }

//____________________________________________________________
void TagEditionDialog::reset()
{
    Debug::Throw( "TagEditionDialog::reset.\n" );
    files_.clear();
    backup_.reset();
    locationEditor_->clear();
    trackNumberEditor_->setEnabled(true);
    _updateEditors();
}

//____________________________________________________________
bool TagEditionDialog::setFiles( const File::List& files )
{

    Debug::Throw( "TagEditionDialog::setFile.\n" );
    reset();
    addFiles( files );
    return readFiles();
}

//____________________________________________________________
void TagEditionDialog::addFiles( const File::List& files )
{
    Debug::Throw( "TagEditionDialog::addFiles.\n" );
    files_.append( files );
}

//____________________________________________________________
bool TagEditionDialog::readFiles()
{

    Debug::Throw( "TagEditionDialog::readFiles.\n" );
    trackNumberEditor_->setEnabled( true );

    // keep track of valid files
    File::List validFiles;

    // keep track of which information is common to all files
    TrackInformation::Mask mask =
        TrackInformation::Artist|
        TrackInformation::Album|
        TrackInformation::Year|
        TrackInformation::Disc|
        TrackInformation::Title|
        TrackInformation::Track;

    // reset writable files flag
    hasWritableFiles_ = false;

    // keep track of location
    File location;
    for( const auto& file:files_ )
    {

        // check taglib interface
        const auto interface = TagLibInterface( file );
        if( !interface.isValid() ) continue;

        // check if file is writable
        if( !hasWritableFiles_ && QFileInfo( file ).isWritable() )
        { hasWritableFiles_ = true; }


        if( validFiles.empty() )
        {

            backup_ = interface.trackInformation();
            location = file;

        } else {

            // check location
            QString first( location.get() );
            QString second( file.get() );

            if( first.length() > second.length() ) first.swap( second );
            auto iter = std::mismatch( first.begin(), first.end(), second.begin() ).first;

            location = File( first.left( iter - first.begin() ) );
            if( !( location.exists() && location.isDirectory() ) ) location = location.path();

            // when there is more than one valid file, track number editor is automatically disabled
            backup_.setTrack(0);
            mask &= ~TrackInformation::Track;
            trackNumberEditor_->setEnabled( false );

            // get track information
            const auto trackInformation( interface.trackInformation( mask ) );
            if( mask & TrackInformation::Artist && trackInformation.artist() != backup_.artist() ) { backup_.setArtist( QString() ); mask &= ~TrackInformation::Artist; }
            if( mask & TrackInformation::Album && trackInformation.album() != backup_.album() ) { backup_.setAlbum( QString() ); mask &= ~TrackInformation::Album; }
            if( mask & TrackInformation::Year && trackInformation.year() != backup_.year() ) { backup_.setYear(0); mask &= ~TrackInformation::Year; }
            if( mask & TrackInformation::Disc && trackInformation.disc() != backup_.disc() ) { backup_.setDisc(0); mask &= ~TrackInformation::Disc; }
            if( mask & TrackInformation::Title && trackInformation.title() != backup_.title() ) { backup_.setTitle( QString() ); mask &= ~TrackInformation::Artist; }

        }

        // append file to internal list
        validFiles.append( file );

    }

    // assign to widgets
    locationEditor_->setText( location.get() );
    _updateEditors();

    // swap sored files with valid
    std::swap( files_, validFiles );
    return !files_.empty();

}

//____________________________________________________________
void TagEditionDialog::setArtists( const QStringList& artists )
{
    Debug::Throw( "TagEditionDialog::setArtists.\n" );
    artistEditor_->clear();
    artistEditor_->addItems( artists );
}

//____________________________________________________________
void TagEditionDialog::setAlbums( const QStringList& albums )
{
    Debug::Throw( "TagEditionDialog::setAlbums.\n" );
    albumEditor_->clear();
    albumEditor_->addItems( albums );
}

//____________________________________________________________
void TagEditionDialog::_updateButtons()
{ okButton().setEnabled( _modified() && hasWritableFiles_ ); }

//____________________________________________________________
bool TagEditionDialog::_modified() const
{
    return
        ( artistEditor_->isEnabled() && artistEditor_->currentText() != backup_.artist() ) ||
        ( albumEditor_->isEnabled() && albumEditor_->currentText() != backup_.album() ) ||
        ( titleEditor_->isEnabled() && titleEditor_->text() != backup_.title() ) ||
        ( yearEditor_->isEnabled() && yearEditor_->value() != backup_.year() && yearEditor_->value() > 0 ) ||
        ( discNumberEditor_->isEnabled() && discNumberEditor_->value() != backup_.disc() && discNumberEditor_->value() > 0 ) ||
        ( trackNumberEditor_->isEnabled() && trackNumberEditor_->value() != backup_.track() && trackNumberEditor_->value() > 0 );
}

//____________________________________________________________
void TagEditionDialog::_saveTags()
{
    Debug::Throw( "TagEditionDialog::_saveTags.\n" );

    if( !_modified() ) return;

    // check modifications
    TrackInformation::Mask mask = TrackInformation::None;
    if( artistEditor_->isEnabled() && artistEditor_->currentText() != backup_.artist() ) {  backup_.setArtist( artistEditor_->currentText() ); mask |= TrackInformation::Artist; }
    if( albumEditor_->isEnabled() && albumEditor_->currentText() != backup_.album() ) {  backup_.setAlbum( albumEditor_->currentText() ); mask |= TrackInformation::Album; }
    if( yearEditor_->isEnabled() && yearEditor_->value() != backup_.year() && yearEditor_->value() > 0 ) {  backup_.setYear( yearEditor_->value() ); mask |= TrackInformation::Year; }
    if( discNumberEditor_->isEnabled() && discNumberEditor_->value() != backup_.disc() && discNumberEditor_->value() > 0 ) {  backup_.setDisc( discNumberEditor_->value() ); mask |= TrackInformation::Disc; }
    if( titleEditor_->isEnabled() && titleEditor_->text() != backup_.title() ) {  backup_.setTitle( titleEditor_->text() ); mask |= TrackInformation::Title; }
    if( trackNumberEditor_->isEnabled() && trackNumberEditor_->value() != backup_.track() && trackNumberEditor_->value() > 0 ) {  backup_.setTrack( trackNumberEditor_->value() ); mask |= TrackInformation::Track; }

    // get taglib interface
    for( const auto file:files_ )
    {

        auto interface = TagLibInterface( file );
        if( !interface.isValid() ) return;

        // assign
        interface.setTrackInformation( backup_, mask );

    }

}

//____________________________________________________________
void TagEditionDialog::_updateEditors()
{

    // update editors from backups
    artistEditor_->setEditText( backup_.artist() );
    albumEditor_->setEditText( backup_.album() );
    titleEditor_->setText( backup_.title() );
    yearEditor_->setValue( std::max( backup_.year(), 0 ) );
    discNumberEditor_->setValue( std::max( backup_.disc(), 0 ) );
    trackNumberEditor_->setValue( std::max( backup_.track(), 0 ) );

    readOnlyLabel_->setVisible( !hasWritableFiles_ );

}

#include "TagEditionDialog.moc"
