
/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "VolumeControlMenu.h"

#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QLayout>
#include <QWidgetAction>

//_________________________________________________________________________
namespace
{
    namespace Local
    {
        class Slider: public QSlider
        {

            Q_OBJECT

            public:

            Slider( QWidget* parent ):
                QSlider( parent )
            {}

            //*@name accessors
            //@{

            QSize minimumSizeHint() const override
            { return QSize( 0, 200 ); }

            //@}

            protected:

            //* wheel event
            void wheelEvent( QWheelEvent* event ) override
            {
                // pass to parent class
                QSlider::wheelEvent( event );

                // always accept so that it is not propagated to parent widgets
                event->accept();
            }
        };

    }
}

//_________________________________________________________________________
VolumeControlMenu::VolumeControlMenu( QWidget* parent ):
    QMenu( parent ),
    Counter( "VolumeControlMenu" )
{
    auto mainWidget = new QWidget( this );
    auto vLayout = new QVBoxLayout( mainWidget );
    mainWidget->setLayout( vLayout );

    vLayout->addWidget( volumeLabel_ = new QLabel( mainWidget ), 0, Qt::AlignCenter );

    const auto fontMetrics( volumeLabel_->fontMetrics() );
    volumeLabel_->setMinimumWidth( fontMetrics.width( tr( "100%" ) ) );
    volumeLabel_->setAlignment( Qt::AlignCenter );
    volumeLabel_->setText( tr( "0%" ) );

    vLayout->addWidget( volumeSlider_ = new Local::Slider( mainWidget ), 1, Qt::AlignCenter );
    volumeSlider_->setOrientation( Qt::Vertical );
    volumeSlider_->setMinimum(0);
    volumeSlider_->setMaximum(100);
    volumeSlider_->setValue(0);
    connect( volumeSlider_, SIGNAL(valueChanged(int)), SLOT(_volumeChanged(int)) );

    vLayout->addWidget( muteButton_ = new QToolButton( mainWidget ), 0, Qt::AlignCenter );
    muteButton_->setIcon( IconEngine::get( IconNames::VolumeMuted ) );
    muteButton_->setText( tr( "Mute" ) );
    muteButton_->setCheckable( true );
    muteButton_->setChecked( false );
    muteButton_->setAutoRaise( true );
    connect( muteButton_, SIGNAL(toggled(bool)), SLOT(_muteStateChanged(bool)) );

    auto sliderActionWidget = new QWidgetAction( this );
    sliderActionWidget->setDefaultWidget( mainWidget );
    addAction( sliderActionWidget );

    // create animation if not already there
    fadeAnimation_ = new QPropertyAnimation( this );
    fadeAnimation_->setPropertyName( "fadeVolume" );
    fadeAnimation_->setTargetObject( this );
    connect( fadeAnimation_, SIGNAL(finished()), this, SLOT(_fadeCompleted()) );

    //* configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________________________________
bool VolumeControlMenu::isFading() const
{ return fadeAnimation_->state() == QPropertyAnimation::Running; }

//_________________________________________________________________________
void VolumeControlMenu::startFade()
{
    Debug::Throw( "VolumeControlMenu::startFade.\n" );

    // do nothing if fade animation is already running
    if( fadeAnimation_->state() == QPropertyAnimation::Running ) return;

    // do nothing if volume is zero, or muted, or if fading is disabled
    if( volumeSlider_->value() == 0 || muteButton_->isChecked() || !useFadeAnimation_ )
    {
        emit fadeCompleted();
        return;
    }

    // store volume
    fadeStartingVolume_ = volumeSlider_->value();
    fadeVolume_ = volumeSlider_->value();

    // setup animation
    fadeAnimation_->setStartValue( fadeStartingVolume_ );
    fadeAnimation_->setEndValue( 0 );
    fadeAnimation_->setDuration( fadeAnimationDuration_ );
    fadeAnimation_->setDirection( QPropertyAnimation::Forward );
    fadeAnimation_->start();

    emit fadeStarted();

}

//_________________________________________________________________________
void VolumeControlMenu::abortFade()
{

    Debug::Throw( "VolumeControlMenu::abortFade.\n" );

    // do nothing if fade animation is not running
    if( fadeAnimation_->state() != QPropertyAnimation::Running ) return;

    // stop animation
    fadeAnimation_->stop();

    // emit abort signal
    emit fadeAborted();

    // restore volume
    emit volumeChanged( fadeStartingVolume_ );

}

//_________________________________________________________________________
void VolumeControlMenu::setFadeVolume( int value )
{
    if( fadeVolume_ == value ) return;
    fadeVolume_ = value;
    emit volumeChanged( fadeVolume_ );
}

//_________________________________________________________________________
void VolumeControlMenu::setVolume( int value )
{
    Debug::Throw() << "VolumeControlMenu::setVolume - volume: " << value << endl;
    volumeLabel_->setText( tr( "%1%" ).arg( value) );
    volumeSlider_->setValue( value );
}

//_________________________________________________________________________
void VolumeControlMenu::wheelEvent( QWheelEvent* event )
{
    event->accept();
    if( !muteButton_->isChecked() )
    { QApplication::sendEvent( volumeSlider_, new QWheelEvent( *event ) ); }
    return;
}

//_________________________________________________________________________
void VolumeControlMenu::_updateConfiguration()
{
    Debug::Throw( "VolumeControlMenu::_updateConfiguration.\n" );
    fadeAnimationDuration_ = XmlOptions::get().get<int>( "FADE_ANIMATION_DURATION" );
    useFadeAnimation_ = XmlOptions::get().get<bool>( "USE_FADE_ANIMATION" ) && fadeAnimationDuration_ > 0;

    if(
        XmlOptions::get().contains( "VOLUME_CONTROL_MUTED" ) &&
        XmlOptions::get().get<bool>( "VOLUME_CONTROL_MUTED" ) &&
        XmlOptions::get().contains( "VOLUME_CONTROL_MUTE_VOLUME" ) )
    {
        lock();
        muteButton_->setChecked( true );
        setVolume( XmlOptions::get().get<int>( "VOLUME_CONTROL_MUTE_VOLUME" ) );
        unlock();
    }

}

//_________________________________________________________________________
void VolumeControlMenu::_volumeChanged( int value )
{
    Debug::Throw() << "VolumeControlMenu::_volumeChanged - volume: " << value << endl;
    volumeLabel_->setText( tr( "%1%" ).arg( value) );

    if( !locked_ ) emit volumeChanged( value );
}

//_________________________________________________________________________
void VolumeControlMenu::_muteStateChanged( bool muted )
{
    Debug::Throw() << "VolumeControlMenu::_muteStateChanged - muted: " << muted << endl;
    volumeLabel_->setEnabled( !muted );
    volumeSlider_->setEnabled( !muted );
    if( !locked_ )
    {

        XmlOptions::get().set<bool>( "VOLUME_CONTROL_MUTED", muted );
        if( muted ) XmlOptions::get().set<int>( "VOLUME_CONTROL_MUTE_VOLUME", volumeSlider_->value() );

        emit muteStateChanged( muted );
        emit volumeChanged( muted ? 0:volumeSlider_->value() );
    }

}

//_________________________________________________________________________
void VolumeControlMenu::_fadeCompleted()
{

    Debug::Throw( "VolumeControlMenu::_fadeCompleted.\n" );

    // emit abort signal
    emit fadeCompleted();

    // restore volume
    emit volumeChanged( fadeStartingVolume_ );

}

#include "VolumeControlMenu.moc"
